/*----------------------------------------------------------------------
  File    : trials.c
  Contents: neuron spike train trial management
  Author  : Christian Borgelt
  History : 2008.11.19 file created
            2008.11.24 most statistics functions implemented
            2008.11.25 bug in function trs_read() fixed (last bin)
            2008.11.27 function trs_crop() added (equalize trial lens.)
            2008.11.28 function trs_cpo() added and optimized
            2009.01.31 parameter 'split' added to trs_read()
            2009.02.18 bug in trial marker skipping fixed
            2009.05.26 function trs_cxc() added
            2009.05.29 functions trs_ccf() and trs_cpo() improved
            2009.06.02 function trs_cpo() improved, trs_csw() added
            2009.06.04 functions trs_ccr(), trs_csr(), tsr_cwr() added
            2011.11.30 normalization with number of positive terms added
            2013.08.09 adapted to direction parameter of int_heapsort()
            2014.04.04 trs_ccr() corrected (divide by y+1 instead of y)
----------------------------------------------------------------------*/
#include <stdio.h>
#include <assert.h>
#include <limits.h>
#include <math.h>
#include "arrays.h"
#include "trials.h"

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#define BLKSIZE     64          /* block size for array enlargement */

/*----------------------------------------------------------------------
  Constants
----------------------------------------------------------------------*/
static TIMEBIN empty[1] = {     /* --- an empty time bin */
  { 0, { -1 } } };              /* (one constant to save memory) */

static TRSTATFN* statfns[] = {  /* --- statistic functions */
  /* TRS_BRE   0 */  trs_bre,   /* background rate estimation */
  /* TRS_CPC   1 */  trs_cpc,   /* conditional pattern complexity */
  /* TRS_CXC   2 */  trs_cxc,   /* conditional excess complexity */
  /* TRS_CCF   3 */  trs_ccf,   /* conditional complexity frequency */
  /* TRS_CCR   4 */  trs_ccr,   /* conditional complexity freq. ratio */
  /* TRS_CFF   5 */  trs_cff,   /* conditional Fano factor */
  /* TRS_CSF   6 */  trs_csf,   /* conditional spike frequency */
  /* TRS_CSF   7 */  trs_csr,   /* conditional spike freq. ratio */
  /* TRS_CSW   8 */  trs_csw,   /* conditional spike weight */
  /* TRS_CSW   9 */  trs_cwr,   /* conditional spike weight ratio */
  /* TRS_CPO  10 */  trs_cpo    /* conditional pattern overlap */
};

/*----------------------------------------------------------------------
  Auxiliary Functions
----------------------------------------------------------------------*/

static double powi (double x, int n)
{                               /* --- raise x to the power of n */
  double r = (n & 1) ? x : 1;   /* result */

  assert(n >= 0);               /* check the function arguments */
  for (n >>= 1; n > 0; n >>= 1){/* traverse the powers of 2 and */
    x *= x; if (n & 1) r *= x; }/* multiply them into the result */
  return r;                     /* if the corr. exponent bit is set, */
}  /* powi() */                 /* finally return the result */

/*--------------------------------------------------------------------*/

static int xbsearch (double key, int *array, int n, int corr)
{                               /* --- do an extended binary search */
  int l, r, m;                  /* array indices */

  assert(array && (n >= 0));    /* check the function arguments */
  for (l = 0, r = n; l < r; ) { /* while search range is not empty */
    m = (l+r) >> 1;             /* compare key to middle element */
    if (key > array[m] +(m+1)*(double)corr) l = m+1;
    else                                    r = m;
  }                             /* adapt the search range accordingly */
  return (l >= n) ? n-1 : l;    /* return the found index */
}  /* xbsearch() */

/*----------------------------------------------------------------------
  Time Bin Functions
----------------------------------------------------------------------*/

TIMEBIN* bin_create (int size)
{                               /* --- create a time bin */
  TIMEBIN *bin;                 /* created time bin */

  assert(size >= 0);            /* check the function argument */
  if (size <= 0) return empty;  /* check for an empty time bin */
  bin = (TIMEBIN*)malloc(sizeof(TIMEBIN) +(size_t)size *sizeof(int));
  if (!bin) return NULL;        /* create a time bin */
  bin->size       = size;       /* note the bin size */
  bin->nids[size] = -1;         /* and store a sentinel */
  return bin;                   /* return the created bin */
}  /* bin_create() */

/*--------------------------------------------------------------------*/

TIMEBIN* bin_createx (int *nids, int n)
{                               /* --- create and init. a time bin */
  TIMEBIN *bin;                 /* created time bin */

  assert((n >= 0) && (nids || (n <= 0)));    /* check the arguments */
  bin = bin_create(n);          /* create a time bin of proper size */
  if (!bin) return NULL;        /* and copy the neuron ids. into it */
  while (--n >= 0) bin->nids[n] = nids[n];
  return bin;                   /* return the created time bin */
}  /* bin_createx() */

/*--------------------------------------------------------------------*/

void bin_delete (TIMEBIN *bin)
{ if (bin != empty) free(bin); }

/*--------------------------------------------------------------------*/

TIMEBIN* bin_filter (TIMEBIN *bin, int *flags)
{                               /* --- filter a time bin */
  int i, k, n;                  /* loop variables, buffer */

  assert(bin && flags);         /* check the function arguments */
  for (i = k = 0; i < bin->size; i++) {
    n = bin->nids[i];           /* traverse the neuron identifiers */
    if (flags[n] != 0) bin->nids[k++] = n;
  }                             /* remove unflagged identifiers */
  if (k <= 0) { bin_delete(bin); return empty; }
  bin->nids[bin->size = k] = -1;/* store a sentinel at the end */
  return (TIMEBIN*)realloc(bin, sizeof(TIMEBIN) +(size_t)k*sizeof(int));
}  /* bin_filter() */           /* resize the time bin */

/*--------------------------------------------------------------------*/
#ifndef NDEBUG

void bin_show (TIMEBIN *bin)
{                               /* --- show neurons in a time bin */
  int i;                        /* loop variable */

  for (i = 0; i < bin->size; i++) {
    if (i > 0) printf(" ");     /* traverse the neuron identifiers, */
    printf("%d", bin->nids[i]); /* print a separator if necessary, */
  }                             /* and then print the neuron id */
  printf("\n");                 /* terminate the neuron list */
}  /* bin_show() */

#endif
/*----------------------------------------------------------------------
  Trial Functions
----------------------------------------------------------------------*/

TRIAL* tr_create (int size)
{                               /* --- create a trial */
  TRIAL *trial;                 /* created trial */

  assert(size >= 0);            /* check the function argument */
  trial = (TRIAL*)malloc(sizeof(TRIAL));
  if (!trial) return NULL;      /* create the base structure and */
  trial->bins = (size <= 0) ? NULL      /* initialize the fields */
    : (TIMEBIN**)malloc((size_t)size *sizeof(TIMEBIN*));
  trial->size = (trial->bins) ? size : 0;
  trial->max  = trial->spc = trial->cnt  = trial->sel = 0;
  trial->isbs = trial->ref = trial->rcxs = NULL;
  return trial;                 /* return the created trial */
}  /* tr_create() */

/*--------------------------------------------------------------------*/

void tr_delete (TRIAL *trial)
{                               /* --- delete a trial */
  assert(trial);                /* check the function argument */
  if (trial->bins) {            /* if there are time bins */
    while (--trial->cnt >= 0)   /* traverse the time bin array */
      bin_delete(trial->bins[trial->cnt]);
  }                             /* delete each time bin */
  if (trial->rcxs) free(trial->rcxs);
  free(trial);                  /* delete the auxiliary arrays */
}  /* tr_delete() */            /* and the base structure */

/*--------------------------------------------------------------------*/

int tr_add (TRIAL *trial, TIMEBIN *bin)
{                               /* --- add a time bin to a trial */
  int     n;                    /* (new) number of array elements */
  TIMEBIN **b;                  /* buffer for reallocation */
  int     *p;                   /* to traverse a time bin */

  assert(trial && bin);         /* check the function arguments */
  n = trial->size;              /* get the current array size */
  if (trial->cnt >= n) {        /* if the time bin array is full */
    n += (n >= BLKSIZE) ? n >> 1 : BLKSIZE;
    b  = (TIMEBIN**)realloc(trial->bins, (size_t)n *sizeof(TIMEBIN*));
    if (!b) return -1;          /* enlarge the time bin array */
    trial->bins = b; trial->size = n;
  }                             /* store the new array and its size */
  trial->bins[trial->cnt] = bin;/* add the new bin to the array */
  trial->spc += bin->size;      /* sum the spikes and update */
  for (p = bin->nids; *p > 0; p++)  /* the maximum neuron id */
    if (*p > trial->max) trial->max = *p;
  return trial->cnt++;          /* return the array index */
}  /* tr_add() */

/*--------------------------------------------------------------------*/

int tr_addx (TRIAL *trial, int *ids, int n)
{                               /* --- add a time bin to a trial */
  TIMEBIN *bin;                 /* created time bin */

  assert(trial                  /* check the function arguments */
  &&    (n >= 0) && (ids || (n <= 0)));
  bin = bin_createx(ids, n);    /* create and add a time bin */
  return (bin) ? tr_add(trial, bin) : -1;
}  /* tr_addx() */

/*--------------------------------------------------------------------*/

int tr_prepare (TRIAL *trial, int nid)
{                               /* --- prepare a trial for a neuron */
  int i, k, m, sum;             /* loop variables, bin size sum */
  int *o, *p, *b, *s;           /* to traverse a bin, buffers */

  assert(trial && (nid > 0));   /* check the function arguments */
  p = (int*)realloc(trial->rcxs, (size_t)trial->cnt *3*sizeof(int));
  if (!p) return -1;            /* allocate memory for arrays */
  trial->rcxs = o = p;          /* and organize it */
  trial->isbs = trial->ref = p += trial->cnt;
  trial->sums = s = p +trial->cnt;
  for (trial->spc = i = k = sum = 0; i < trial->cnt; i++) {
    trial->spc += m = trial->bins[i]->size;
    s[i] = sum += m;            /* traverse the time bins */
    o[i] = 0; p[i] = i;         /* and sum their sizes */
    for (b = trial->bins[i]->nids; *b > 0; b++) {
      if (*b != nid) o[i]++;    /* traverse the neuron ids */
      else { p[i] = p[k]; p[k++] = i; }
    }                           /* count spikes of other neurons and */
  }                             /* collect bin ids for given neuron */
  return trial->sel = k;        /* return the number of time bins */
}  /* tr_prepare() */           /* in which the given neuron fires */

/*--------------------------------------------------------------------*/

void tr_shuffle (TRIAL *trial, double corr, RANDFN *randfn)
{                               /* --- shuffle the spikes of a neuron */
  int    i, k, n;               /* index, loop variable, buffers */
  int    *s, *o;                /* array of selected bin indices */
  double z;                     /* range for random numbers */

  assert(trial && randfn);      /* check the function arguments */
  if (corr < 0) {               /* shuffle the bin indices uniformly */
    int_select(trial->isbs, (size_t)trial->cnt,
              (size_t)trial->sel, randfn); return; }
  o = trial->rcxs; s = trial->isbs;
  n = trial->cnt;               /* get arrays, number of time bins, */
  z = trial->spc +corr*n;       /* and the range for random numbers */
  if (trial->sel > (n << 1)) {  /* if complement smaller than sel. */
    for (k = n-trial->sel; --k >= 0; ) {
      do { i = xbsearch(z *randfn(), trial->sums, n, (int)corr);
      } while (o[i] < 0);       /* find position of a random number */
      o[i] |= INT_MIN;          /* and check whether bin is marked */
    }                           /* if not, mark the selected bin */
    for (i = k = 0; i < n; i++) {
      if (o[i] < 0) o[i] &= ~INT_MIN;
      else          s[k++] = i; /* collect indices of unmarked bins */
    } }                         /* and unmark all marked bins */
  else {                        /* if selection smaller than compl. */
    for (k = trial->sel; --k >= 0; ) {
      do { i = xbsearch(z *randfn(), trial->sums, n, (int)corr);
      } while (o[i] < 0);       /* find position of random number */
      o[s[k] = i] |= INT_MIN;   /* mark the selected bin */
    }                           /* and store its index */
    for (k = trial->sel; --k >= 0; )
      o[s[k]] &= ~INT_MIN;      /* unmark the selected bins */
  }                             /* (restore original state) */
}  /* tr_shuffle() */

/*--------------------------------------------------------------------*/
#ifndef NDEBUG

void tr_show (TRIAL *trial)
{                               /* --- show the time bins of a trial */
  int i;                        /* loop variable */

  for (i = 0; i < trial->cnt; i++) {
    printf("%3d: ", i);         /* traverse the time bins, */
    bin_show(trial->bins[i]);   /* print a bin identifier and */
  }                             /* then the neuron identifiers */
}  /* tr_show() */

#endif
/*----------------------------------------------------------------------
  Trial Set Functions
----------------------------------------------------------------------*/

TRIALSET* trs_create (int size)
{                               /* --- create a trial set */
  TRIALSET *trset;              /* created trial set */

  assert(size >= 0);            /* check the function argument */
  trset = (TRIALSET*)malloc(sizeof(TRIALSET));
  if (!trset) return NULL;      /* create the base structure and */
  if (size <= 0) trset->trials = NULL;          /* a trial array */
  else trset->trials = (TRIAL**)malloc((size_t)size *sizeof(TRIAL*));
  trset->size = (trset->trials) ? size : 0;
  trset->max  = trset->cnt  = trset->sum = trset->sel = trset->mcx = 0;
  trset->mrgs = trset->frqs = NULL;    /* initialize the fields */
  trset->code = trset->parm = -1;
  return trset;                 /* return the created trial set */
}  /* trs_create() */

/*--------------------------------------------------------------------*/

void trs_delete (TRIALSET *trset)
{                               /* --- delete a trial set */
  assert(trset);                /* check the function argument */
  if (trset->trials) {          /* if there are trials */
    while (--trset->cnt >= 0)   /* traverse the trials */
      tr_delete(trset->trials[trset->cnt]);
  }                             /* delete each trial */
  if (trset->mrgs) free(trset->mrgs);
  free(trset);                  /* delete the auxiliary arrays */
}  /* trs_delete() */           /* and the base structure */

/*--------------------------------------------------------------------*/

static int error (int code, TRIAL *trial, int *nids)
{                               /* --- clean up after an error */
  if (trial) tr_delete(trial);  /* delete an existing trial */
  if (nids)  free(nids);        /* and an existing buffer */
  return code;                  /* return the error code */
}  /* error() */

/*--------------------------------------------------------------------*/

int trs_read (TRIALSET *trset, FILE *file,
              int split, double offset, double delta)
{                               /* --- read a trial from a file */
  int    *nids  = NULL, *p;     /* buffer for one time bin */
  TRIAL  *trial = NULL;         /* created trial */
  int    k;                     /* number of fields */
  int    n, size;               /* number of neurons, array size */
  int    read;                  /* flag for gap between trials */
  double id, time;              /* neuron identifier, spike time */
  double start, end;            /* start time and end of next bin */

  assert(trset && file          /* check the function arguments */
  &&    (offset >= 0) && (delta > 0));
  start = offset;               /* default: base time is 0 */
  end   = start +delta;         /* get start time and end of time bin */
  size  = n = read = 0;         /* neuron read buffer is empty */
  while (1) {                   /* trial read loop */
    k = fscanf(file, "%lg %lg\n", &id, &time);
    if (k != 2) {               /* read next identifier/time pair */
      if (k != EOF) return error(-2, trial, nids);
      if (!trial)   break;      /* check for a read error and */
      id = -1; time = end;      /* for a regular end of file */
    }                           /* set a dummy (end) marker */
    if ((int)id != id)          /* check the read identifier */
      return error(-3, trial, nids);
    if (id <= 0)                /* clarify the marker meaning */
      id = (read || (k == EOF)) ? -1 : 0;
    if (id == 0) {              /* if start of trial marker, */
      start = time  +offset;    /* get new time values */
      end   = start +delta; n = 0; continue;
    }                           /* skip the entry */
    if ((id >= 0) && !trial) {  /* if there is no active trial */
      trial = tr_create(BLKSIZE);
      if (!trial) error(-1, NULL, nids);
      read = -1;                /* start a new trial and */
    }                           /* set the read flag */
    if ((id > 0) && (time < start))
      continue;                 /* skip spikes before start time */
    while (time > end) {        /* while outside of current time bin */
      int_heapsort(nids, (size_t)n, +1);    /* sort neuron ids and */
      n = (int)int_unique(nids, (size_t)n);   /* remove duplicates */
      if (tr_addx(trial, nids, n) < 0) /* add the bin to the trial */
        return error(-1, trial, nids);
      n = 0; end += delta;      /* get end of the next time bin */
    }                           /* and reinit. the neuron counter */
    if (id < 0) {               /* if end of trial marker or EOF */
      if (time >= end) {        /* if last bin has not been stored */
        int_heapsort(nids, (size_t)n, +1);  /* sort neuron ids and */
        n = (int)int_unique(nids, (size_t)n); /* remove duplicates */
        if (tr_addx(trial, nids, n) < 0)
          return error(-1, trial, nids);
      }                         /* add the bin to the trial */
      start = time  +offset;    /* get new time values (in case */
      end   = start +delta;     /* next start marker is missing) */
      read  = n = 0;            /* reinit. the spike counter */
      if (split || (k == EOF)){ /* if to split into trials */
        if (trs_add(trset, trial) < 0)
          return error(-1, trial, nids);
        trial = NULL;           /* add active trial to the trial set */
      } }                       /* and then clear the trial variable */
    else if (id > 0) {          /* if id is a neuron identifier */
      if (n >= size) {          /* if the neuron array is full */
        size += (size > BLKSIZE) ? size >> 1 : BLKSIZE;
        p = realloc(nids, (size_t)size *sizeof(int));
        if (!p) return error(-1, trial, nids);
        nids = p;               /* enlarge the neuron array */
      }                         /* and set the new array */
      nids[n++] = (int)id;      /* store the neuron identifier and */
    }                           /* increment the neuron counter */
    if (k == EOF) break;        /* check for end of input */
  }
  if (nids) free(nids);         /* delete the neuron read buffer */
  return 0;                     /* return 'ok' */
}  /* trs_read() */

/*--------------------------------------------------------------------*/

int trs_add (TRIALSET *trset, TRIAL *trial)
{                               /* --- add a trial to a trial set */
  int   n;                      /* (new) number of array elements */
  TRIAL **p;                    /* buffer for reallocation */

  assert(trset && trial);       /* check the function arguments */
  n = trset->size;              /* get the current array size */
  if (trset->cnt >= n) {        /* if the trset array is full */
    n += (n >= BLKSIZE) ? n >> 1 : BLKSIZE;
    p  = (TRIAL**)realloc(trset->trials, (size_t)n *sizeof(TRIAL*));
    if (!p) return -1;          /* enlarge the trial array */
    trset->trials = p; trset->size = n;
  }                             /* store the new array and its size */
  trset->trials[trset->cnt] = trial;
  if (trial->max > trset->max)  /* add the trial to the array and */
    trset->max = trial->max;    /* update the maximum neuron id */
  trial->spc += trial->spc;     /* sum the number of spikes */
  trset->sum += trial->cnt;     /* and the number of time bins */
  trset->code = trset->parm = -1;
  trset->nid  = -1;             /* clear statistics parameters */
  return trset->cnt++;          /* return the trial array index */
}  /* trs_add() */

/*--------------------------------------------------------------------*/

int trs_diff (TRIALSET *trset)
{                               /* --- check trials for equal size */
  int   i;                      /* loop variable */
  TRIAL **t;                    /* to access the trials */

  assert(trset);                /* check the function argument */
  if (trset->cnt < 2) return -1;/* check for more than one trial */
  t = trset->trials;            /* traverse the trials */
  for (i = 0; i < trset->cnt; i++)
    if (t[i]->cnt != t[0]->cnt) return t[i]->cnt;
  return -1;                    /* check for equal trial lengths */
}  /* trs_diff() */

/*--------------------------------------------------------------------*/

int trs_crop (TRIALSET *trset, int cnt)
{                               /* --- crop all trials of the set */
  int   i, k;                   /* loop variable, buffer */
  TRIAL *t;                     /* to traverse the trials */

  assert(trset);                /* check the function arguments */
  if (cnt < 0) cnt = INT_MAX;   /* check and adapt the trial length */
  for (i = trset->cnt; --i >= 0; ) {
    k = trset->trials[i]->cnt;  /* traverse the trials of the set */
    if (k < cnt) cnt = k;       /* find length of shortest trial */
  }                             /* (or use given trial length) */
  for (i = trset->cnt; --i >= 0; ) {
    t = trset->trials[i];       /* traverse the trials of the set */
    while (t->cnt > cnt)        /* traverse excess time bins */
      t->spc -= t->bins[--t->cnt]->size;
  }                             /* crop the trials to equal length */
  trset->sum  = cnt *trset->cnt;
  trset->code = trset->parm = -1;
  trset->nid  = -1;             /* clear the statistic parameters */
  return cnt;                   /* return the new trial length */
}  /* trs_crop() */

/*--------------------------------------------------------------------*/

int trs_filter (TRIALSET *trset, int min)
{                               /* --- remove neurons with few spikes */
  int   i, k, n;                /* loop variables, neuron counter */
  TRIAL *t;                     /* to traverse the trials */
  int   *flags;                 /* flags for neurons */

  assert(trset && (min >= 0));  /* check the function arguments */
  trs_prepare(trset, 1);        /* prepare for an arbitrary neuron */
  if (min <= 1) return trset->max;
  trs_csf(trset, 1, 0);         /* count the spikes per neuron */
  flags = trset->frqs;          /* get a flag array for the neurons */
  for (n = 0, i = trset->max+1; --i >= 0; ) {
    if (trset->mrgs[i] >= min) { flags[i] = 1; n++; }
    else {      trset->mrgs[i] = flags[i] = 0; }
  }                             /* set or clear flags, count neurons */
  for (k = trset->cnt; --k >= 0; ) {
    t = trset->trials[k];       /* traverse the trials */
    for (i = t->cnt; --i >= 0;) /* traverse the time bins */
      t->bins[i] = bin_filter(t->bins[i], flags);
  }                             /* filter the time bins */
  return n;                     /* return new number of neurons */
}  /* trs_filter() */

/*--------------------------------------------------------------------*/

int trs_prepare (TRIALSET *trset, int nid)
{                               /* --- prepare trial set for a neuron */
  int   i, k;                   /* loop variables */
  TRIAL *t;                     /* to traverse the trials */
  int   *p;                     /* buffer for reallocation */

  assert(trset && (nid > 0));   /* check the function arguments */
  trset->max = trset->spc = trset->sum = trset->sel = trset->mcx = 0;
  for (i = trset->cnt; --i >= 0; ) {
    t = trset->trials[i];       /* traverse and prepare the trials */
    if (tr_prepare(t, nid) < 0) return -1;
    trset->spc += t->spc;       /* sum the number of spikes */
    trset->sum += t->cnt;       /* the number of time bins, */
    trset->sel += t->sel;       /* and the number of selected bins */
    if (t->max > trset->max) trset->max = t->max;
  }                             /* find maximum neuron identifier */
  k = (trset->max > 0) ? trset->max : 1;
  trset->avg = ((trset->spc > 0) ? trset->spc : 1) /(double)k;
  k = trset->max +1;            /* compute the array sizes */
  p = (int*)realloc(trset->mrgs, (size_t)(k+k) *sizeof(int));
  if (!p) return -1;            /* enlarge the frequency arrays, */
  trset->mrgs = p;              /* store the new arrays, and */
  trset->frqs = p += k;         /* clear the marginals array */
  trset->nid  = nid;            /* note neuron and clear parameters */
  trset->code = trset->parm = -1;
  return trset->sel;            /* return the number of time bins */
}  /* trs_prepare() */          /* in which the given neuron fires */

/*--------------------------------------------------------------------*/

void trs_shuffle (TRIALSET *trset, double corr, RANDFN *randfn)
{                               /* --- shuffle a trial set */
  int   i, k, n, x;             /* loop variables, exchange buffer */
  TRIAL **t;                    /* to access the trial array */
  int   *p;                     /* exchange buffer */

  assert(trset && randfn);      /* check the function arguments */
  n = trset->cnt; t = trset->trials;
  if (n <= 0) return;           /* check for no or only one trial */
  if (n <= 1) { tr_shuffle(t[0], corr, randfn); return; }
  for (k = n; k > 0; ) {        /* shuffle loop (n-1 selections) */
    i = (int)(randfn() *k--);   /* compute a random index */
    if (i > k) i = k;           /* in the remaining section and */
    if (i < 0) i = 0;           /* clamp it to a valid range */
    p = t[i]->isbs; t[i]->isbs = t[k]->isbs; t[k]->isbs = p;
    x = t[i]->sel;  t[i]->sel  = t[k]->sel;  t[k]->sel  = x;
  }                             /* exchange the index arrays */
  for (k = n; --k >= 0; ) {     /* avoid unshuffled elements */
    if (t[k]->isbs != t[k]->ref) continue;
    i = (int)(randfn() *n);     /* compute a random index */
    if (i > n) i = n;           /* in the whole array and */
    if (i < 0) i = 0;           /* clamp it to a valid range */
    p = t[i]->isbs; t[i]->isbs = t[k]->isbs; t[k]->isbs = p;
    x = t[i]->sel;  t[i]->sel  = t[k]->sel;  t[k]->sel  = x;
  }                             /* exchange the index arrays */
}  /* trs_shuffle() */

/*--------------------------------------------------------------------*/

double trs_bre (TRIALSET *trset, int max, int dummy)
{                               /* --- base rate estimation */
  int    i, k, n;               /* loop variables, buffer */
  TRIAL  *t;                    /* to traverse the trials */
  double eta, xi;               /* firing probabilities */

  assert(trset && (max >= 0));  /* check the function arguments */
  if (trset->sel < 1) return 0; /* check for at least one spike */
  if ((trset->code != TRS_BRE)  /* if this is the first time */
  ||  (trset->parm != max)) {   /* this statistic is computed, */
    trset->code = TRS_BRE;      /* note the code of the statistic */
    trset->parm = max;          /* and its parameter */
    for (n = 0, k = trset->cnt; --k >= 0; ) {
      t = trset->trials[k];     /* traverse the trials */
      for (i = t->cnt; --i >= 0; )
        if (t->rcxs[i] <= max) n++;
    }                           /* count bins with at most max spikes */
    trset->buf = n;             /* and note this number */
  }                             /* (for background rate estimation) */
  if (trset->buf <= 0) return 0;
  for (n = 0, k = trset->cnt; --k >= 0; ) {
    t = trset->trials[k];       /* traverse the trials */
    for (i = t->sel; --i >= 0;) /* traverse the selected bins */
      if (t->rcxs[t->isbs[i]] <= max) n++;
  }                             /* count bins with at most max spikes */
  eta = trset->sel /(double)trset->sum;
  xi  = (n  > 0) ? n /trset->buf : 0;
  xi  = (xi < 1) ? (eta -xi) /(1.0 -xi) : 0;
  xi /= eta;                    /* compute (rel.) firing probability */
  return xi;                    /* return the value of the statistic */
}  /* trs_bre() */

/*--------------------------------------------------------------------*/

double trs_cpc (TRIALSET *trset, int expo, int dummy)
{                               /* --- cond. pattern complexity */
  int    i, k;                  /* loop variables */
  TRIAL  *t;                    /* to traverse the trials */
  double sum;                   /* sum of powers of complexities */

  assert(trset && (expo > 0));  /* check the function arguments */
  if (trset->sel < 1) return 0; /* check for at least one spike */
  if ((trset->code != TRS_CPC)  /* if this is the first time */
  ||  (trset->parm != expo)) {  /* this statistic is computed, */
    trset->code = TRS_CPC;      /* note the code of the statistic */
    trset->parm = expo;         /* and its parameter */
    for (sum = 0, k = trset->cnt; --k >= 0; ) {
      t = trset->trials[k];     /* traverse the trials */
      for (i = t->cnt; --i >= 0; )
        sum += powi(t->rcxs[i], expo);
    }                           /* sum powers of reduced complexities */
    trset->buf = sum/trset->sum;/* compute and note the overall */
  }                             /* average pattern complexity */
  for (sum = 0, k = trset->cnt; --k >= 0; ) {
    t = trset->trials[k];       /* traverse the trials */
    for (i = t->sel; --i >= 0; )
      sum += powi(t->rcxs[t->isbs[i]], expo);
  }                             /* sum powers for selected bins */
  sum = (sum/trset->sel -trset->buf)
      / trset->buf;             /* compute relative complexity diff. */
  return sum;                   /* return the value of the statistic */
}  /* trs_cpc() */

/*--------------------------------------------------------------------*/

double trs_cxc (TRIALSET *trset, int expo, int npos)
{                               /* --- cond. excess complexity */
  int    i, k, n;               /* loop variables */
  TRIAL  *t;                    /* to traverse the trials */
  double sum;                   /* sum of powers of complexities */
  double x;                     /* temporary buffer */

  assert(trset && (expo > 0));  /* check the function arguments */
  if (trset->sel < 1) return 0; /* check for at least one spike */
  if ((trset->code != TRS_CXC)  /* if this is the first time */
  ||  (trset->parm != expo)) {  /* this statistic is computed, */
    trset->code = TRS_CXC;      /* note the code of the statistic */
    trset->parm = expo;         /* and its parameter */
    trset->buf  = (trset->spc -trset->sel) /(double)trset->sum;
  }                             /* compute average pattern complexity */
  for (sum = n = 0, k = trset->cnt; --k >= 0; ) {
    t = trset->trials[k];       /* traverse the trials */
    for (i = t->sel; --i >= 0; ) {
      x = t->rcxs[t->isbs[i]] -trset->buf;
      if (x > 0) { sum += powi(x, expo); n++; }
    }                           /* sum the requested powers */
  }                             /* of the excess complexities */
  if (!npos) n = trset->sel;    /* get the number of time bins */
  if (n > 1) sum /= n;          /* normalize the computed sum */
  return sum;                   /* return the value of the statistic */
}  /* trs_cxc() */

/*--------------------------------------------------------------------*/

static void cxfrq (TRIALSET *trset, int cond)
{                               /* --- det. complexity frequencies */
  int   i, k;                   /* loop variables */
  TRIAL *t;                     /* to traverse the trials */

  if (cond) {                   /* if conditional frequencies */
    for (i = trset->max+1; --i >= 0; )
      trset->frqs[i] = 0;       /* clear the complexity frequencies */
    for (k = trset->cnt; --k >= 0; ) {
      t = trset->trials[k];     /* traverse the trials */
      for (i = t->sel; --i >= 0; )
        trset->frqs[t->rcxs[t->isbs[i]]] += 1;
    } }                         /* count the complexity occurrences */
  else {                        /* if marginal frequencies */
    for (i = trset->max+1; --i >= 0; )
      trset->mrgs[i] = 0;       /* clear the marginal frequencies */
    for (k = trset->cnt; --k >= 0; ) {
      t = trset->trials[k];     /* traverse the trials */
      for (i = t->cnt; --i >= 0; )  /* traverse the time bins */
        trset->mrgs[t->rcxs[i]] += 1;
    }                           /* count the complexity occurrences */
    for (trset->mcx = i = 0; i <= trset->max; i++)
      if (trset->mrgs[i] > 0) trset->mcx = i;
  }                             /* find the maximal complexity */
}  /* cxfrq() */

/*--------------------------------------------------------------------*/

double trs_ccf (TRIALSET *trset, int expo, int dummy)
{                               /* --- cond. complexity frequencies */
  int    k;                     /* loop variable */
  double sum;                   /* sum of powers of frequencies */
  double x, r, n;               /* temporary buffers */

  assert(trset && (expo > 0));  /* check the function arguments */
  if (trset->sel < 1) return 0; /* check for at least one spike */
  if ((trset->code != TRS_CCF)  /* if this is the first time */
  ||  (trset->parm != expo)) {  /* this statistic is computed, */
    trset->code = TRS_CCF;      /* note the code of the statistic */
    trset->parm = expo;         /* and its parameter */
    cxfrq(trset, 0);            /* determine the marginal */
  }                             /* complexity frequencies */
  cxfrq(trset, 1);              /* det. cond. complexity frequencies */
  r = trset->sel/(double)trset->sum;
  n = (trset->spc -trset->sel) /(double)trset->sum;
  for (sum = 0, k = (int)ceil(n); k <= trset->mcx; k++) {
    x = trset->frqs[k] -r *trset->mrgs[k];
    if (x > 0) sum += x *powi(k-n, expo);
  }                             /* sum the excess frequencies */
  if (trset->max > 1) sum /= trset->max;
  return sum;                   /* return the value of the statistic */
}  /* trs_ccf() */

/*--------------------------------------------------------------------*/

double trs_ccr (TRIALSET *trset, int expo, int dummy)
{                               /* --- cond. complexity freq. ratio */
  int    k;                     /* loop variable */
  double n, sum;                /* sum of powers of frequencies */
  double x, y, r;               /* temporary buffers */

  assert(trset && (expo > 0));  /* check the function arguments */
  if (trset->sel < 1) return 0; /* check for at least one spike */
  if ((trset->code != TRS_CCR)  /* if this is the first time */
  ||  (trset->parm != expo)) {  /* this statistic is computed, */
    trset->code = TRS_CCR;      /* note the code of the statistic */
    trset->parm = expo;         /* and its parameter */
    cxfrq(trset, 0);            /* determine the marginal */
  }                             /* complexity frequencies */
  cxfrq(trset, 1);              /* det. cond. complexity frequencies */
  r = trset->sel/(double)trset->sum;
  n = (trset->spc -trset->sel) /(double)trset->sum;
  for (sum = 0, k = (int)ceil(n); k <= trset->mcx; k++) {
    x = trset->frqs[k] -(y = r *trset->mrgs[k]);
    if (x > 0) sum += x/(y+1) *powi((k-n), expo);
  }                             /* sum the excess frequency ratios */
  if (trset->max > 1) sum /= trset->max;
  return sum;                   /* return the value of the statistic */
}  /* trs_ccr() */

/*--------------------------------------------------------------------*/

double trs_cff (TRIALSET *trset, int expo, int dummy)
{                               /* --- conditional Fano factor */
  int    i, k;                  /* loop variables */
  TRIAL  *t;                    /* to traverse the trials */
  double sum;                   /* sum of powers of complexities */
  double m;                     /* temporary buffer */

  assert(trset && (expo > 0));  /* check the function arguments */
  if (trset->sel < 2) return 0; /* check for at least two spikes */
  if ((trset->code != TRS_CFF)  /* if this is the first time */
  ||  (trset->parm != expo)) {  /* this statistic is computed, */
    trset->code = TRS_CFF;      /* note the code of the statistic */
    trset->parm = expo;         /* and its parameter */
    m = (trset->spc -trset->sel) /(double)trset->sum;
    for (sum = 0, k = trset->cnt; --k >= 0; ) {
      t = trset->trials[k];     /* traverse the trials */
      for (i = t->cnt; --i >= 0; )
        sum += powi(fabs(t->rcxs[i]-m), expo+1);
    }                           /* sum powers of complexity diffs. */
    trset->buf = (m > 0) ? sum /(trset->sum-1) /m : 0;
  }                             /* compute unconditional Fano factor */
  for (m = 0, k = trset->cnt; --k >= 0; ) {
    t = trset->trials[k];       /* traverse the trials */
    for (i = t->sel; --i >= 0; ) m += t->rcxs[t->isbs[i]];
  }                             /* sum complexities of selected bins */
  m /= trset->sel;              /* compute cond. mean complexity */
  for (sum = 0, k = trset->cnt; --k >= 0; ) {
    t = trset->trials[k];       /* traverse the trials */
    for (i = t->sel; --i >= 0; )
      sum += powi(fabs(t->rcxs[t->isbs[i]] -m), expo+1);
  }                             /* sum powers for selected bins */
  sum = (m > 0) ? sum /(trset->sel-1) /m : 0;
  sum  = (sum -trset->buf)      /* compute conditional Fano factor */
           /   trset->buf;      /* and relate it ot the uncond. one */
  return sum;                   /* return the value of the statistic */
}  /* trs_cff() */

/*--------------------------------------------------------------------*/

static void spfrq (TRIALSET *trset, int cond)
{                               /* --- determine spike frequencies */
  int   i, k;                   /* loop variables */
  TRIAL *t;                     /* to traverse the trials */
  int   *b;                     /* to traverse the time bins */

  if (cond) {                   /* if conditional frequencies */
    for (i = trset->max+1; --i >= 0; )
      trset->frqs[i] = 0;       /* clear the marginal frequencies */
    for (k = trset->cnt; --k >= 0; ) {
      t = trset->trials[k];     /* traverse the trials */
      for (i = t->sel; --i >= 0; )
        for (b = t->bins[t->isbs[i]]->nids; *b > 0; b++)
          trset->frqs[*b] += 1; /* traverse the time bins and */
    } }                         /* count the spikes per neuron */
  else {                        /* if marginal frequencies */
    for (i = trset->max+1; --i >= 0; )
      trset->mrgs[i] = 0;       /* clear the marginal frequencies */
    for (k = trset->cnt; --k >= 0; ) {
      t = trset->trials[k];     /* traverse the trials */
      for (i = t->cnt; --i >= 0; )
        for (b = t->bins[i]->nids; *b > 0; b++)
          trset->mrgs[*b] += 1; /* traverse the time bins and */
    }                           /* count the spikes per neuron */
  }
  trset->frqs[trset->nid] = 0;  /* do not consider current neuron */
}  /* spfrq() */

/*--------------------------------------------------------------------*/

double trs_csf (TRIALSET *trset, int expo, int npos)
{                               /* --- conditional spike frequencies */
  int    k, n;                  /* loop variables */
  double sum;                   /* sum of powers of frequencies */
  double x, r;                  /* temporary buffers */

  assert(trset && (expo > 0));  /* check the function arguments */
  if (trset->sel < 1) return 0; /* check for at least one spike */
  if ((trset->code != TRS_CSF)  /* if this is the first time */
  ||  (trset->parm != expo)) {  /* this statistic is computed, */
    trset->code = TRS_CSF;      /* note the code of the statistic */
    trset->parm = expo;         /* and its parameter */
    spfrq(trset, 0);            /* determine the marginal */
  }                             /* marginal spike frequencies */
  spfrq(trset, 1);              /* det. conditional spike frequencies */
  r = trset->sel/(double)trset->sum;
  for (sum = n = 0, k = trset->max+1; --k >= 0; ) {
    x = trset->frqs[k] -r *trset->mrgs[k];
    if (x > 0) { sum += powi(x, expo); n++; }
  }                             /* sum the excess frequencies */
  if (!npos) n = trset->max-1;  /* get the number of neurons */
  if (n > 1) sum /= n;          /* normalize the computed sum */
  return sum;                   /* return the value of the statistic */
}  /* trs_csf() */

/*--------------------------------------------------------------------*/

double trs_csr (TRIALSET *trset, int expo, int npos)
{                               /* --- conditional spike freq. ratio */
  int    k, n;                  /* loop variables */
  double sum;                   /* sum of powers of frequencies */
  double x, y, r;               /* temporary buffers */

  assert(trset && (expo > 0));  /* check the function arguments */
  if (trset->sel < 1) return 0; /* check for at least one spike */
  if ((trset->code != TRS_CSR)  /* if this is the first time */
  ||  (trset->parm != expo)) {  /* this statistic is computed, */
    trset->code = TRS_CSR;      /* note the code of the statistic */
    trset->parm = expo;         /* and its parameter */
    spfrq(trset, 0);            /* determine the marginal */
  }                             /* marginal spike frequencies */
  spfrq(trset, 1);              /* det. conditional spike frequencies */
  r = trset->sel/(double)trset->sum;
  for (sum = n = 0, k = trset->max+1; --k >= 0; ) {
    x = trset->frqs[k] -(y = r *trset->mrgs[k]);
    if (x > 0) { sum += powi(x/y, expo); n++; }
  }                             /* sum the excess frequency ratios */
  if (!npos) n = trset->max-1;  /* get the number of neurons */
  if (n > 1) sum /= n;          /* normalize the computed sum */
  return sum;                   /* return the value of the statistic */
}  /* trs_csr() */

/*--------------------------------------------------------------------*/

static void spwgt (TRIALSET *trset, int cond)
{                               /* --- determine spike weights */
  int    i, k, n, c;            /* loop variables, buffers */
  TRIAL  *t;                    /* to traverse the trials */
  int    *b;                    /* to traverse the neuron ids */

  if (cond) {                   /* if conditional weights */
    for (i = trset->max+1; --i >= 0; )
      trset->frqs[i] = 0;       /* clear the marginal frequencies */
    for (k = trset->cnt; --k >= 0; ) {
      t = trset->trials[k];     /* traverse the trials */
      for (i = t->sel; --i >= 0; ) {
        n = t->isbs[i];         /* traverse the time bins */
        if ((c = t->rcxs[n]) < 2) continue;
        for (b = t->bins[n]->nids; *b > 0; b++)
          trset->frqs[*b] += c; /* traverse the time bins and */
      }                         /* sum the weights per neuron */
    } }                         /* (conditional spike weights) */
  else {                        /* if marginal weights */
    for (i = trset->max+1; --i >= 0; )
      trset->mrgs[i] = 0;       /* clear the marginal frequencies */
    for (k = trset->cnt; --k >= 0; ) {
      t = trset->trials[k];     /* traverse the trials */
      for (i = t->cnt; --i >= 0; ) {
        if ((c = t->rcxs[i]) < 2) continue;
        for (b = t->bins[i]->nids; *b > 0; b++)
          trset->mrgs[*b] += c; /* traverse the time bins and */
      }                         /* sum the weights per neuron */
    }                           /* (marginal spike weights) */
  }
  trset->frqs[trset->nid] = 0;  /* do not consider current neuron */
}  /* spwgt() */

/*--------------------------------------------------------------------*/

double trs_csw (TRIALSET *trset, int expo, int npos)
{                               /* --- conditional spike weight */
  int    k, n;                  /* loop variables */
  double sum;                   /* sum of powers of frequencies */
  double x, r;                  /* temporary buffers */

  assert(trset && (expo > 0));  /* check the function arguments */
  if (trset->sel < 1) return 0; /* check for at least one spike */
  if ((trset->code != TRS_CSW)  /* if this is the first time */
  ||  (trset->parm != expo)) {  /* this statistic is computed, */
    trset->code = TRS_CSW;      /* note the code of the statistic */
    trset->parm = expo;         /* and its parameter */
    spwgt(trset, 0);            /* determine the marginal */
  }                             /* spike weights */
  spwgt(trset, 1);              /* det. conditional spike weights */
  r = trset->sel/(double)trset->sum;
  for (sum = n = 0, k = trset->max+1; --k >= 0; ) {
    x = trset->frqs[k] -r *trset->mrgs[k];
    if (x > 0) { sum += powi(x, expo); n++; }
  }                             /* sum the excess spike weights */
  if (!npos) n = trset->max-1;  /* get the number of neurons */
  if (n > 1) sum /= n;          /* normalize the computed sum */
  return sum;                   /* return the value of the statistic */
}  /* trs_csw() */

/*--------------------------------------------------------------------*/

double trs_cwr (TRIALSET *trset, int expo, int npos)
{                               /* --- conditional spike weight ratio */
  int    k, n;                  /* loop variables */
  double sum;                   /* sum of powers of frequencies */
  double x, y, r;               /* temporary buffers */

  assert(trset && (expo > 0));  /* check the function arguments */
  if (trset->sel < 1) return 0; /* check for at least one spike */
  if ((trset->code != TRS_CWR)  /* if this is the first time */
  ||  (trset->parm != expo)) {  /* this statistic is computed, */
    trset->code = TRS_CWR;      /* note the code of the statistic */
    trset->parm = expo;         /* and its parameter */
    spwgt(trset, 0);            /* determine the marginal */
  }                             /* spike weights */
  spwgt(trset, 1);              /* det. conditional spike weights */
  r = trset->sel/(double)trset->sum;
  for (sum = n = 0, k = trset->max+1; --k >= 0; ) {
    x = trset->frqs[k] -(y = r *trset->mrgs[k]);
    if (x > 0) { sum += powi(x/y, expo); n++; }
  }                             /* sum the excess weight ratios */
  if (!npos) n = trset->max-1;  /* get the number of neurons */
  if (n > 1) sum /= n;          /* normalize the computed sum */
  return sum;                   /* return the value of the statistic */
}  /* trs_cwr() */

/*--------------------------------------------------------------------*/

double trs_cpo (TRIALSET *trset, int expo, int dummy)
{                               /* --- conditional pattern overlap */
  int     i, j, k, l, n, x;     /* loop variables, buffer */
  TRIAL   *t, *u;               /* to traverse the trials */
  TIMEBIN *a, *b;               /* to traverse the time bins */
  double  sum;                  /* sum of pattern overlaps */
  int     *flags;               /* neuron containment flags */

  assert(trset && (expo > 0));  /* check the function arguments */
  if (trset->sel < 2) return 0; /* check for at least two spikes */
  if ((trset->code != TRS_CPO)  /* if this is the first time */
  ||  (trset->parm != expo)) {  /* this statistic is computed, */
    trset->code = TRS_CPO;      /* note the code of the statistic */
    trset->parm = expo;         /* and its parameter */
  }                             /* (no common values to compute) */
  flags = trset->frqs;          /* clear the containment flag array */
  for (n = 0; n <= trset->max; n++) flags[n] = 0;
  for (sum = 0, k = trset->cnt; --k >= 0; ) {
    t = trset->trials[k];       /* traverse the trials */
    for (i = t->sel; --i >= 0; ) {
      a = t->bins[t->isbs[i]];  /* traverse the time bins */
      if (a->size <= 1) continue;
      if ((a->size <= 2)        /* if potential overlap is <= 1 */
      && ((a->nids[0] == trset->nid)
      ||  (a->nids[1] == trset->nid)))
        continue;               /* skip the time bin */
      /* Checking the reduced pattern complexity with */
      /*   n = t->isbs[i];                            */
      /*   if (t->rcxs[n] <= 1) continue;             */
      /*   a = t->bins[n];                            */
      /* is equivalent, but slightly less efficient.  */
      for (n = a->size; --n >= 0; )
        flags[a->nids[n]] = 1;  /* set flags for contained neurons */
      flags[trset->nid] = 0;    /* clear flag for current neuron */
      for (l = k+1; --l >= 0; ) {
        u = trset->trials[l];   /* traverse the remaining trials */
        for (j = (l == k) ? i : u->sel; --j >= 0; ) {
          b = u->bins[u->isbs[j]];    /* traverse the time bins, but */
          if (b->size <= 1) continue; /* skip bin if overlap is <= 1 */
          for (x = 0, n = b->size; --n >= 0; )
            x += flags[b->nids[n]];
          if (x > 1) sum += powi(x, expo);
        }                       /* traverse the time bins and */
      }                         /* compute the pattern overlap */
      for (n = a->size; --n >= 0; )
        flags[a->nids[n]] = 0;  /* clear flags for contained neurons */
    }                           /* (faster than clearing full array) */
  }                             /* afterwards compute average overlap */
  sum /= 0.5 *trset->sel *(double)(trset->sel-1);
  /* sum *= sqrt(trset->sum/trset->sel); */
  return sum;                   /* return the value of the statistic */
}  /* trs_cpo() */

/*--------------------------------------------------------------------*/
#ifndef NDEBUG

void trs_show (TRIALSET *trset)
{                               /* --- show the trials in a set */
  int i;                        /* loop variable */

  for (i = 0; i < trset->cnt; i++) {
    printf("trial %d\n", i);    /* traverse all trials, */
    tr_show(trset->trials[i]);  /* print the trial identifier, */
  }                             /* and then the trial */
}  /* tr_show() */

#endif
/*--------------------------------------------------------------------*/

TRSTATFN* trs_statfn (int code)
{ return statfns[code]; }
