/*----------------------------------------------------------------------
  File    : trials.h
  Contents: neuron spike train trial management
  Author  : Christian Borgelt
  History : 2008.11.19 file created
            2009.01.31 parameter 'split' added to trs_read()
            2009.05.26 function trs_cxc() added
            2009.06.02 function trs_csw() added
            2009.06.04 functions trs_ccr(), trs_csr(), tsr_cwr() added
            2011.11.30 parameter npos added to statistic functions
----------------------------------------------------------------------*/
#ifndef __TRIALS__
#define __TRIALS__
#include <stdlib.h>
#include "fntypes.h"

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
/* --- statistics codes --- */
#define TRS_BRE     0           /* background rate estimation */
#define TRS_CPC     1           /* conditional pattern complexity */
#define TRS_CXC     2           /* conditional excess complexity */
#define TRS_CCF     3           /* conditional complexity frequency */
#define TRS_CCR     4           /* conditional complexity freq. ratio */
#define TRS_CFF     5           /* conditional Fano factor */
#define TRS_CSF     6           /* conditional spike frequency */
#define TRS_CSR     7           /* conditional spike freq. ratio */
#define TRS_CSW     8           /* conditional spike weight */
#define TRS_CWR     9           /* conditional spike weight ratio */
#define TRS_CPO    10           /* conditional pattern overlap */

/*----------------------------------------------------------------------
  Type Definitions
----------------------------------------------------------------------*/
typedef struct {                /* --- a time bin --- */
  int     size;                 /* number of spikes/neurons */
  int     nids[1];              /* neuron identifiers */
} TIMEBIN;                      /* (time bin) */

typedef struct {                /* --- a trial --- */
  int     max;                  /* maximum neuron identifier */
  int     spc;                  /* total spike count */
  int     cnt;                  /* current number of bins */
  int     size;                 /* current array size */
  TIMEBIN **bins;               /* array of time bins */
  int     *rcxs;                /* reduced pattern complexities */
  int     sel;                  /* number  of selected bins */
  int     *isbs;                /* indices of selected bins */
  int     *ref;                 /* reference for trial shuffling */
  int     *sums;                /* sums of bins sizes */
} TRIAL;                        /* (trial) */

typedef struct {                /* --- a set of trials --- */
  int     cnt;                  /* current number of trials */
  int     size;                 /* current array size */
  TRIAL   **trials;             /* array of trials */
  int     max;                  /* maximum neuron identifier */
  int     spc;                  /* total spike count */
  int     sum;                  /* sum of the numbers of time bins */
  int     sel;                  /* number of selected time bins */
  int     mcx;                  /* maximum pattern complexity */
  double  avg;                  /* average number of spikes */
  int     *frqs;                /* spike/complexity frequencies */
  int     *mrgs;                /* ditto, marginal frequencies */
  int     nid;                  /* identifier of current neuron */
  int     code;                 /* code of last test statistic */
  int     parm;                 /* last test statistic parameter */
  double  buf;                  /* buffer for auxiliary value */
} TRIALSET;                     /* (trial set) */

typedef double TRSTATFN (TRIALSET* trset, int param, int npos);

/*----------------------------------------------------------------------
  Time Bin Functions
----------------------------------------------------------------------*/
extern TIMEBIN*  bin_create  (int size);
extern TIMEBIN*  bin_createx (int *nids, int n);
extern void      bin_delete  (TIMEBIN *bin);

extern int       bin_size    (TIMEBIN *bin);
extern int*      bin_nids    (TIMEBIN *bin);

extern TIMEBIN*  bin_filter  (TIMEBIN *bin, int *flags);

#ifndef NDEBUG
extern void      bin_show    (TIMEBIN *bin);
#endif

/*----------------------------------------------------------------------
  Trial Functions
----------------------------------------------------------------------*/
extern TRIAL*    tr_create   (int size);
extern void      tr_delete   (TRIAL *trial);

extern int       tr_max      (TRIAL *trial);
extern int       tr_cnt      (TRIAL *trial);
extern int       tr_add      (TRIAL *trial, TIMEBIN *bin);
extern int       tr_addx     (TRIAL *trial, int *ids, int n);
extern TIMEBIN*  tr_get      (TRIAL *trial, int index);

extern int       tr_prepare  (TRIAL *trial, int nid);
extern void      tr_shuffle  (TRIAL *trial, double corr,
                              RANDFN *randfn);
#ifndef NDEBUG
extern void      tr_show     (TRIAL *trial);
#endif

/*----------------------------------------------------------------------
  Trial Set Functions
----------------------------------------------------------------------*/
extern TRIALSET* trs_create  (int size);
extern void      trs_delete  (TRIALSET *trset);
extern int       trs_read    (TRIALSET *trset, FILE *file,
                              int split, double offset, double delta);

extern int       trs_max     (TRIALSET *trset);
extern int       trs_cnt     (TRIALSET *trset);
extern int       trs_add     (TRIALSET *trset, TRIAL *trial);
extern int       trs_get     (TRIALSET *trset, int index);

extern int       trs_diff    (TRIALSET *trset);
extern int       trs_crop    (TRIALSET *trset, int cnt);
extern int       trs_filter  (TRIALSET *trset, int min);

extern int       trs_prepare (TRIALSET *trset, int nid);
extern void      trs_shuffle (TRIALSET *trset, double corr,
                              RANDFN *randfn);

extern double    trs_bre     (TRIALSET *trset, int max,  int dummy);
extern double    trs_cpc     (TRIALSET *trset, int expo, int dummy);
extern double    trs_cxc     (TRIALSET *trset, int expo, int npos);
extern double    trs_ccf     (TRIALSET *trset, int expo, int dummy);
extern double    trs_ccr     (TRIALSET *trset, int expo, int dummy);
extern double    trs_cff     (TRIALSET *trset, int expo, int dummy);
extern double    trs_csf     (TRIALSET *trset, int expo, int npos);
extern double    trs_csr     (TRIALSET *trset, int expo, int npos);
extern double    trs_csw     (TRIALSET *trset, int expo, int npos);
extern double    trs_cwr     (TRIALSET *trset, int expo, int npos);
extern double    trs_cpo     (TRIALSET *trset, int expo, int dummy);

#ifndef NDEBUG
extern void      trs_show    (TRIALSET *trset);
#endif

extern TRSTATFN* trs_statfn  (int code);

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#define bin_size(b)         ((b)->size)
#define bin_nids(b)         ((b)->nids)

/*--------------------------------------------------------------------*/
#define tr_max(t)           ((t)->max)
#define tr_cnt(t)           ((t)->cnt)
#define tr_get(t,i)         ((t)->bins[i])

/*--------------------------------------------------------------------*/
#define trs_max(s)          ((s)->max)
#define trs_cnt(s)          ((s)->cnt)
#define trs_get(s,i)        ((s)->trials[i])

#endif
